using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace Shisen_Sho {
    public partial class Board : UserControl {
        private Graphics graphics, offscr_g;
        private Bitmap offscrBitmap;
        private int num_tiles_h;
        private int num_tiles_v;
        private Grid grid;
        private Point selectedPt;
        private Point clickedPt;
        private Point[] path;
        //private Pen penHighlight;
        private Brush brushHighlight;
        private Pen pathPen;
        private Pen pathErasePen;

        private bool gameStarted = false;
        private int gameTime = 0;
        private bool cheated = false;

        private const int TILE_WIDTH = 40;
        private const int TILE_HEIGHT = 56;
        private const int TILE_MARGIN = 0;
        private const int BOARD_MARGIN = 20;

        private const float PATH_WIDTH = 6;
        private const int HIGHLIGHT_TRANSPARENCY = 50;

        public EventHandler gameWon;
        public EventHandler gameLost;
        public EventHandler cheatedChanged;

        public static Tile[] pieces = {
            new Tile("bamboo"),
            new Tile("bird"),
            new Tile("char_1"),
            new Tile("char_2"),
            new Tile("char_3"),
            new Tile("char_4"),
            new Tile("char_5"),
            new Tile("char_6"),
            new Tile("char_7"),
            new Tile("char_8"),
            new Tile("char_9"),
            new Tile("circle_2"),
            new Tile("circle_3"),
            new Tile("circle_4"),
            new Tile("circle_5"),
            new Tile("circle_6"),
            new Tile("circle_7"),
            new Tile("circle_8"),
            new Tile("circle_9"),
            new Tile("dragon_1"),
            new Tile("dragon_2"),
            new Tile("dragon_3"),
            new Tile("mum"),
            new Tile("orchid"),
            new Tile("plum"),
            new Tile("stick_2"),
            new Tile("stick_3"),
            new Tile("stick_4"),
            new Tile("stick_5"),
            new Tile("stick_6"),
            new Tile("stick_7"),
            new Tile("stick_8"),
            new Tile("stick_9"),
            new Tile("sun"),
            new Tile("tree_empty"),
            new Tile("tree_full")
        };

        public Board() {
            InitializeComponent();
        }

        public event EventHandler OnGameWon {
            add {
                gameWon += value;
            }
            remove {
                gameWon -= value;
            }
        }

        public event EventHandler OnGameLost {
            add {
                gameLost += value;
            }
            remove {
                gameLost -= value;
            }
        }

        public event EventHandler OnCheatedChanged {
            add {
                cheatedChanged += value;
            }
            remove {
                cheatedChanged -= value;
            }
        }

        public void init(int num_tiles_v, int num_tiles_h) {
            setDimensions(num_tiles_v, num_tiles_h);
            grid.reset();
            //penHighlight = new Pen(Color.FromArgb(100, Color.YellowGreen));
            brushHighlight = new SolidBrush(Color.FromArgb(HIGHLIGHT_TRANSPARENCY, Color.Blue));
            pathPen = new Pen(Color.Purple, PATH_WIDTH);
            pathErasePen = new Pen(Color.White, PATH_WIDTH);
            selectedPt = Point.Empty;

            //num_tiles_h = (this.Width - 2 * BOARD_MARGIN - TILE_MARGIN) / (TILE_WIDTH + TILE_MARGIN);
            //num_tiles_v = (this.Height - 2 * BOARD_MARGIN - TILE_MARGIN) / (TILE_HEIGHT + TILE_MARGIN);
            //if (num_tiles_h % 2 == 1 && num_tiles_v % 2 == 1)
            //    num_tiles_h--;

            this.BackColor = Color.White;
            paintBoard();
        }

        public void setDimensions(int num_tiles_v, int num_tiles_h) {
            if (num_tiles_v != this.num_tiles_v || num_tiles_h != this.num_tiles_h) {
                this.num_tiles_h = num_tiles_h;
                this.num_tiles_v = num_tiles_v;
                this.Width = BOARD_MARGIN * 2 + TILE_MARGIN + (TILE_WIDTH + TILE_MARGIN) * num_tiles_h;
                this.Height = BOARD_MARGIN * 2 + TILE_MARGIN + (TILE_HEIGHT + TILE_MARGIN) * num_tiles_v;
                grid = new Grid(num_tiles_v, num_tiles_h);
                graphics = this.CreateGraphics();
                graphics.Clear(this.BackColor);
                offscrBitmap = new Bitmap(this.Width, this.Height);
                offscr_g = Graphics.FromImage(offscrBitmap);
            }
        }

        public void reset() {
            grid.reset();
            gameStarted = false;
            tmrGame.Enabled = false;
            setCheated(false);
            selectedPt = clickedPt = Point.Empty;
            gameTime = 0;
            erasePath();
            paintBoard();
            copyBoard();
        }

        public bool isCheated() {
            return cheated;
        }

        private void setCheated(bool cheated) {
            this.cheated = cheated;
            cheatedChanged(this, EventArgs.Empty);
        }

        public bool isGameStarted() {
            return gameStarted;
        }

        public int getGameTime() {
            return gameTime;
        }

        public int incGameTime() {
            return ++gameTime;
        }

        private void paintBoard(Rectangle paintRect) {
            for (int i = 1; i <= num_tiles_v; i++) {
                for (int j = 1; j <= num_tiles_h; j++) {
                    drawTile(i, j);
                }
            }
        }

        private void paintBoard() {
            paintBoard(Rectangle.Empty);
        }

        private void copyBoard(Rectangle paintRect) {
            if (offscrBitmap != null) {
                if (paintRect != Rectangle.Empty) {
                    graphics.DrawImage(offscrBitmap, paintRect, paintRect, GraphicsUnit.Pixel);
                } else {
                    graphics.DrawImage(offscrBitmap, 0, 0);
                }
            }
        }

        private void copyBoard() {
            copyBoard(Rectangle.Empty);
        }

        /* the tile highlight rect */
        private Rectangle getTileHRect(int y, int x) {
            return new Rectangle(BOARD_MARGIN + TILE_MARGIN / 2 + (x - 1) * (TILE_WIDTH + TILE_MARGIN),
                BOARD_MARGIN + TILE_MARGIN / 2 + (y - 1) * (TILE_HEIGHT + TILE_MARGIN),
                TILE_WIDTH + TILE_MARGIN,
                TILE_HEIGHT + TILE_MARGIN);
        }

        /* the tile paint location */
        private Point getTileLoc(int y, int x) {
            return new Point(BOARD_MARGIN + TILE_MARGIN + (x - 1) * (TILE_WIDTH + TILE_MARGIN),
                BOARD_MARGIN + TILE_MARGIN + (y - 1) * (TILE_HEIGHT + TILE_MARGIN));
        }

        private Point getTileCenter(int y, int x) {
            Point tileLoc = getTileLoc(y, x);
            tileLoc.Y += TILE_HEIGHT / 2;
            tileLoc.X += TILE_WIDTH / 2;
            if (y <= 0 || x <= 0 || y > num_tiles_v || x > num_tiles_h) {
                if (y == 0)
                    tileLoc.Y = BOARD_MARGIN / 2;
                if (x == 0)
                    tileLoc.X = BOARD_MARGIN / 2;
                if (y == num_tiles_v + 1)
                    tileLoc.Y = BOARD_MARGIN + TILE_MARGIN + (TILE_MARGIN + TILE_HEIGHT) * num_tiles_v + BOARD_MARGIN / 2;
                if (x == num_tiles_h + 1)
                    tileLoc.X = BOARD_MARGIN + TILE_MARGIN + (TILE_MARGIN + TILE_WIDTH) * num_tiles_h + BOARD_MARGIN / 2;
            }
            return tileLoc;
        }

        private Point getTileCenter(Point pt) {
            return getTileCenter(pt.Y, pt.X);
        }

        protected override void OnPaint(PaintEventArgs e) {
            base.OnPaint(e);
            copyBoard(e.ClipRectangle);
        }

        private void drawTile(int y, int x) {
            Point loc = getTileLoc(y, x);
            BoardCell cell = grid.getCell(y, x);
            Rectangle tileHRect = getTileHRect(y, x);
            eraseTile(y, x);
            offscr_g.DrawImage(pieces[cell.getType()].getImage(), loc);
            if (cell.isHighlighted()) {
                offscr_g.FillRectangle(brushHighlight, tileHRect);
            }
        }

        private void drawTile(Point pt) {
            drawTile(pt.Y, pt.X);
        }

        private void eraseTile(int y, int x) {
            Rectangle tileHRect = getTileHRect(y, x);
            offscr_g.FillRectangle(Brushes.White, tileHRect);
        }

        private void eraseTile(Point pt) {
            eraseTile(pt.Y, pt.X);
        }

        private void updateTiles(LinkedList<Point> liChanged) {
            foreach (Point pt in liChanged) {
                if (grid.isCellEmpty(pt))
                    eraseTile(pt);
                else
                    drawTile(pt);
            }
        }

        public bool drawHint() {
            setCheated(true);
            if (tmrPath.Enabled)
                return true;
            erasePath();
            Point[] pts;
            if ((pts = grid.getHint()) != null) {
                path = grid.getPath(pts[0], pts[1]);
                selectedPt = Point.Empty;
                LinkedList<Point> llChanged = grid.unHighlightAll();
                updateTiles(llChanged);
                grid.highlightCell(path[0]);
                grid.highlightCell(path[path.Length - 1]);
                drawPath();
                copyBoard();
                return true;
            }
            return false;
        }

        private void Board_MouseDown(object sender, MouseEventArgs e) {
            int tile_x = 1 + (e.X - BOARD_MARGIN - TILE_MARGIN) / (TILE_WIDTH + TILE_MARGIN);
            int tile_y = 1 + (e.Y - BOARD_MARGIN - TILE_MARGIN) / (TILE_HEIGHT + TILE_MARGIN);
            tile_x = tile_x > num_tiles_h ? -1 : tile_x;
            tile_y = tile_y > num_tiles_v ? -1 : tile_y;

            if (tile_x >= 0 && tile_y >= 0 && e.X >= BOARD_MARGIN && e.Y >= BOARD_MARGIN
                && !tmrPath.Enabled && !grid.isCellEmpty(tile_y, tile_x)) {
                clickedPt = new Point(tile_x, tile_y);
                erasePath();
                //Console.WriteLine("Clicked on tile (" + tile_x + ", " + tile_y + ").");
                LinkedList<Point> llChanged = new LinkedList<Point>();
                if (e.Button == MouseButtons.Right) {
                    selectedPt = Point.Empty;
                    llChanged = grid.changeHighlightType(clickedPt);
                } else {
                    if (selectedPt != Point.Empty) {
                        if (selectedPt == clickedPt) {
                            grid.unHighlightCell(selectedPt);
                            drawTile(selectedPt);
                            selectedPt = Point.Empty;
                        } else if (grid.getCell(clickedPt).getType() == grid.getCell(selectedPt).getType()) {
                            path = grid.getPath(selectedPt, clickedPt);
                            if (path != null) {
                                grid.highlightCell(selectedPt);
                                grid.highlightCell(clickedPt);
                                drawPath();
                                tmrPath.Enabled = true;
                                tmrGame.Enabled = true;
                                gameStarted = true;
                            }
                        }
                    } else {
                        llChanged = grid.unHighlightAll();
                        llChanged.AddFirst(clickedPt);
                        grid.highlightCell(clickedPt);
                        selectedPt = clickedPt;
                    }
                }
                updateTiles(llChanged);
                copyBoard();
            }
        }

        private void drawPath() {
            drawTile(path[0]);
            drawTile(path[path.Length - 1]);
            drawPathLine(pathPen);
        }

        private void erasePath() {
            if (path != null) {
                drawPathLine(pathErasePen);
                drawTile(path[0]);
                drawTile(path[path.Length - 1]);
                path = null;
            }
        }

        private void drawPathLine(Pen pen) {
            Point center2 = getTileCenter(path[0]);
            Point center1;
            for (int i = 0; i < path.Length - 1; i++) {
                center1 = center2;
                center2 = getTileCenter(path[i + 1]);
                offscr_g.DrawLine(pen, center1, center2);
            }
        }

        private void drawGameWon() {
            string gameWon = "Game Won!";
            string yourTime = "Your time: " + getGameTime();
            string youCheated = "(but you cheated)";
            Font gwFont = new Font("Verdana", 30, FontStyle.Bold);
            Font ytFont = new Font("Verdana", 25, FontStyle.Bold);
            Font ycFont = new Font("Verdana", 15, FontStyle.Bold);
            SizeF gwSize = graphics.MeasureString(gameWon, gwFont);
            SizeF ytSize = graphics.MeasureString(yourTime, ytFont);
            SizeF ycSize = graphics.MeasureString(youCheated, ycFont);
            PointF gwLoc =  new PointF((this.Width - gwSize.Width - (isCheated() ? ycSize.Width : 0)) / 2, this.Height / 2 - gwSize.Height);
            PointF ytLoc =  new PointF((this.Width - ytSize.Width) / 2, this.Height / 2);
            PointF ycLoc = new PointF(gwLoc.X + gwSize.Width , gwLoc.Y + ycSize.Height/2);

            offscr_g.DrawString(gameWon, gwFont, Brushes.Navy, gwLoc);
            offscr_g.DrawString(yourTime, ytFont, Brushes.Orange, ytLoc);
            if(isCheated())
                offscr_g.DrawString(youCheated, ycFont, Brushes.Navy, ycLoc);
            copyBoard();
        }

        private void tmrPath_Tick(object sender, EventArgs e) {
            tmrPath.Enabled = false;
            erasePath();
            LinkedList<Point> llChanged = grid.clearPair(selectedPt, clickedPt);
            selectedPt = Point.Empty;
            clickedPt = Point.Empty;
            updateTiles(llChanged);
            copyBoard();
            if (grid.isGameWon()) {
                tmrGame.Enabled = false;
                gameStarted = false;
                drawGameWon();
                gameWon(this, EventArgs.Empty);
            } else if (grid.getHint() == null) {
                gameStarted = false;
                tmrGame.Enabled = false;
                gameLost(this, EventArgs.Empty);
            }
        }
    }
}
